"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createAsyncSubscribable = void 0;
/**
 * Create a AsyncSubscribable instance that uses the give subscribe and
 * unsubscribe function to manage listeners.
 *
 * @param options -
 *   options used to create this async subscribable
 * @returns
 *   instance of `AsyncSubscribable`
 */
function createAsyncSubscribable(options) {
    const subscribe = options.subscribe;
    const unsubscribe = options.unsubscribe;
    const subscribable = async (listener) => {
        await subscribe(listener);
        // Return a handle that can be used to unsubscribe
        return {
            async unsubscribe() {
                await unsubscribe(listener);
            }
        };
    };
    subscribable.subscribe = subscribable;
    subscribable.unsubscribe = unsubscribe;
    subscribable.once = () => new Promise((resolve, reject) => {
        const listener = (...args) => {
            unsubscribe(listener)
                .catch(err => reject(err));
            resolve(args);
        };
        subscribe(listener)
            .catch(err => reject(err));
    });
    subscribable.filter = (filter) => createFilteredAsyncSubscribable(subscribe, unsubscribe, filter);
    subscribable.withThis = (newThis) => createNewThisAsyncSubscribable(subscribe, unsubscribe, newThis);
    return subscribable;
}
exports.createAsyncSubscribable = createAsyncSubscribable;
/**
 * Create a AsyncSubscribable that is filtered via the specified function.
 *
 * @param subscribe -
 *   function used to subscribe listeners
 * @param unsubscribe -
 *   function used to unsubscribe listeners
 * @param filterToApply -
 *   the function used to filter events
 * @returns
 *   `AsyncSubscribable`
 */
function createFilteredAsyncSubscribable(subscribe, unsubscribe, filterToApply) {
    // Map used to keep track of the filtered listener of an added listener
    const listenerMapping = new Map();
    return createAsyncSubscribable({
        subscribe: listener => {
            const actualListener = function (...args) {
                if (filterToApply.apply(this, args)) {
                    listener.call(this, ...args);
                }
            };
            listenerMapping.set(listener, actualListener);
            return subscribe(actualListener);
        },
        unsubscribe: listener => {
            const actual = listenerMapping.get(listener);
            if (actual) {
                listenerMapping.delete(listener);
                return unsubscribe(actual);
            }
            else {
                return Promise.resolve(false);
            }
        }
    });
}
/**
 * Create a AsyncSubscribable that changes what this is used for listeners.
 *
 * @param subscribe -
 *   function used to subscribe listeners
 * @param unsubscribe -
 *   function used to unsubscribe listeners
 * @param newThis -
 *   object to use as the new this
 * @returns
 *   `AsyncSubscribable`
 */
function createNewThisAsyncSubscribable(subscribe, unsubscribe, newThis) {
    // Map used to keep track of the modified listeners
    const listenerMapping = new Map();
    return createAsyncSubscribable({
        subscribe: listener => {
            const actualListener = function (...args) {
                listener.call(newThis, ...args);
            };
            listenerMapping.set(listener, actualListener);
            return subscribe(actualListener);
        },
        unsubscribe: listener => {
            const actual = listenerMapping.get(listener);
            if (actual) {
                listenerMapping.delete(listener);
                return unsubscribe(actual);
            }
            else {
                return Promise.resolve(false);
            }
        }
    });
}
//# sourceMappingURL=createAsyncSubscribable.js.map