"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSubscribable = void 0;
/**
 * Create a Subscribable given a subscribe, unsubscribe and a once function.
 *
 * @param options -
 *   options used to create this subscribable
 * @returns
 *   instance of `Subscribable`
 */
function createSubscribable(options) {
    const subscribe = options.subscribe;
    const unsubscribe = options.unsubscribe;
    const subscribable = (listener) => {
        subscribe(listener);
        // Return a handle that can be used to unsubscribe
        return {
            unsubscribe() {
                unsubscribe(listener);
            }
        };
    };
    subscribable.subscribe = subscribable;
    subscribable.unsubscribe = unsubscribe;
    subscribable.once = () => new Promise(resolve => {
        const listener = (...args) => {
            unsubscribe(listener);
            resolve(args);
        };
        subscribe(listener);
    });
    subscribable.filter = (filter) => createFilteredSubscribable(subscribe, unsubscribe, filter);
    subscribable.withThis = (newThis) => createNewThisSubscribable(subscribe, unsubscribe, newThis);
    return subscribable;
}
exports.createSubscribable = createSubscribable;
/**
 * Create a Subscribable that is filtered via the specified function.
 *
 * @param subscribe -
 *   function used to subscribe listeners
 * @param unsubscribe -
 *   function used to unsubscribe listeners
 * @param filterToApply -
 *   function used to fitler events
 * @returns
 *   `Subscribable`
 */
function createFilteredSubscribable(subscribe, unsubscribe, filterToApply) {
    // Map used to keep track of the filtered listener of an added listener
    const listenerMapping = new Map();
    return createSubscribable({
        subscribe: listener => {
            const actualListener = function (...args) {
                if (filterToApply.apply(this, args)) {
                    listener.call(this, ...args);
                }
            };
            listenerMapping.set(listener, actualListener);
            subscribe(actualListener);
        },
        unsubscribe: listener => {
            const actual = listenerMapping.get(listener);
            if (actual) {
                listenerMapping.delete(listener);
                return unsubscribe(actual);
            }
            else {
                return false;
            }
        }
    });
}
/**
 * Create a Subscribable that changes what this is used for listeners.
 *
 * @param subscribe -
 *   function used to subscribe listeners
 * @param unsubscribe -
 *   function used to unsubscribe listeners
 * @param newThis -
 *   object to use as the new this
 * @returns
 *   `Subscribable`
 */
function createNewThisSubscribable(subscribe, unsubscribe, newThis) {
    // Map used to keep track of the modified listeners
    const listenerMapping = new Map();
    return createSubscribable({
        subscribe: listener => {
            const actualListener = function (...args) {
                listener.call(newThis, ...args);
            };
            listenerMapping.set(listener, actualListener);
            subscribe(actualListener);
        },
        unsubscribe: listener => {
            const actual = listenerMapping.get(listener);
            if (actual) {
                listenerMapping.delete(listener);
                return unsubscribe(actual);
            }
            else {
                return false;
            }
        }
    });
}
//# sourceMappingURL=createSubscribable.js.map