import { AsyncListener } from './AsyncListener';
import { AsyncSubscribable } from './AsyncSubscribable';
import { AsyncSubscriptionFunctions } from './AsyncSubscriptionFunctions';
import { AsyncSubscriptionHandle } from './AsyncSubscriptionHandle';
/**
 * An event that handles subscription in an asynchronous way. This type of
 *
 *
 * Each instance represents a single event:
 *
 * ```javascript
 * const event = new AsyncEvent(valueForThisInListeners);
 * ```
 *
 * The event can be emitted via the `emit` method:
 *
 * ```javascript
 * await event.emit('first-param', 'second-param');
 * ```
 *
 * Listeners can be added directly on the event, but it is recommended to
 * use `.subscribable` for any API that is public:
 *
 * ```javascript
 * // Adding a listener directly on the event
 * await event.subscribe(() => ...);
 *
 * // Subscribable provides a public API
 * await event.subscribable(() => ..)
 * await event.subscribable.subscribe(() => ...);
 * ```
 *
 * Listeners can be unsubscribed either via their handle or via the event:
 *
 * ```javascript
 * // Use handle for easier unsubscription
 * const handle = await event.addListener(() => ...);
 * await handle.unsubscribe();
 *
 * // Unsubscribe the actual listener
 * const listener = () => ...;
 * await eventOrSubscribable.subscribe(listener);
 * await eventOrSubscribable.unsubscribe(listener);
 * ```
 */
export declare class AsyncEvent<Parent, Args extends any[] = []> implements AsyncSubscriptionFunctions<Parent, Args> {
    /**
     * Public AsyncSubscribable that can safely be shared with consumers that
     * should be able to listen for events.
     */
    readonly subscribable: AsyncSubscribable<Parent, Args>;
    /**
     * Parent of this handler, used to apply the correct this to event listeners.
     */
    private readonly parent;
    /**
     * Listener(s) that have been attached to this event handler.
     */
    private registeredListeners?;
    /**
     * Monitor that will be notified on any listener change.
     */
    private monitor?;
    /**
     * Create a new event.
     *
     * @param parent -
     *   the parent that will be passed to listener as their `this`
     */
    constructor(parent: Parent);
    /**
     * Emit this event. This will invoke all of the listeners with the passed
     * arguments.
     *
     * @param args -
     *   arguments that listeners will receive
     */
    emit(...args: Args): Promise<void>;
    /**
     * Subscribe to this event using the given listener. The listener will
     * be invoked any time the event is emitted.
     *
     * @param listener -
     *   listener to subscribe
     * @returns
     *   promise
     */
    protected subscribe0(listener: AsyncListener<Parent, Args>): Promise<void>;
    /**
     * Unsubscribe a listener from this handler. The specified listener will
     * no longer be invoked when the event is emitted.
     *
     * @param listener -
     *   listener to unsubscribe
     * @returns
     *   promise indicating if the listener was subscribed
     */
    protected unsubscribe0(listener: AsyncListener<Parent, Args>): Promise<boolean>;
    /**
     * Subscribe to this event using the given listener. The listener will
     * be invoked any time the event is emitted. The returned handle can be
     * used to unsubscribe.
     *
     * @param listener -
     *   listener to subscribe
     * @returns
     *   handle to the subscription, can be used to unsubscribe. Resolves
     *   when the subscription is fully registered
     */
    subscribe(listener: AsyncListener<Parent, Args>): Promise<AsyncSubscriptionHandle>;
    /**
     * Unsubscribe a listener from this handler. The specified listener will
     * no longer be invoked when the event is emitted.
     *
     * @param listener -
     *   listener to unsubscribe
     * @returns
     *   promise that resolves when the listener is fully unsubscribed
     */
    unsubscribe(listener: AsyncListener<Parent, Args>): Promise<boolean>;
    /**
     * Get a promise that will resolve the first time this event is fired
     * after this call.
     *
     * @returns
     *   listener that resolves the next time the event is emitted
     */
    once(): Promise<Args>;
    /**
     * Create a subscribable that will apply the specified filter to any
     * listeners added.
     *
     * @param filter -
     *   function used to filter events
     * @returns
     *   filtered `AsyncSubscribable`
     */
    filter(filter: (...args: Args) => boolean | Promise<boolean>): AsyncSubscribable<Parent, Args>;
    /**
     * Create a Subscribable that changes the this argument used for listeners.
     *
     * @param newThis -
     *   what should be treated as this for event listeners
     * @returns
     *   modified `AsyncSubscribable`
     */
    withThis<NewThis>(newThis: NewThis): AsyncSubscribable<NewThis, Args>;
    /**
     * Get if there are any listeners available.
     *
     * @returns
     *   `true` if listeners are present
     */
    get hasListeners(): boolean;
    /**
     * Get a copy of the listeners as an array.
     *
     * @returns
     *   listeners as array
     */
    get listeners(): AsyncListener<Parent, Args>[];
    /**
     * Clear all listeners for this event.
     */
    clear(): void;
    /**
     * Monitor for changes to listeners. Only a single monitor is supported at
     * a single time. This is intended to be used to react to if listeners are
     * currently registered. This can be used for things such as only listening
     * to events from other objects when this event is active.
     *
     * @param monitor -
     *   function used to monitor for changes to listeners
     */
    monitorListeners(monitor: (event: this) => void): void;
    /**
     * Stop monitoring for listener changes.
     */
    removeMonitor(): void;
}
//# sourceMappingURL=AsyncEvent.d.ts.map